#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2023/12/07 12:58:54
@Author  :   ChenHao
@Contact :   jerrychen1990@gmail.com
'''
import copy
from xagents.model.core import LLM, EMBD
from xagents.model.zhipu import GLM, ZhipuEmbedding
from xagents.model.local import SN_EMBD, SN_LLM

